/*
 *  Copyright (C) (2010-2012) Vadim Biktashev, Alexander Karpov, 
 *                            Irina Biktasheva, Ross McFarlane;
 *  Copyright (C) 2012 Sanjay Kharche
 *
 *  This file is part of Beatbox.
 *
 *   Beatbox is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Beatbox is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Beatbox.  If not, see <http://www.gnu.org/licenses/>.
 */


/* MAIN PROGRAM */

/* these two files are in the parent (trunk) directory */
#include <config.h>
/* #include <subversion.h>  */

#if MPI
#include <mpi.h>
#endif

#include <setjmp.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#define MAIN
#include "system.h"
#include "beatbox.h"
#include "init.h"
#include "device.h"
#include "state.h"
#include "bikt.h"
#include "windraw.h"
#include "qpp.h"
#include "mpi_io_choice.h"

/* Exported variables */

/* options */
#if MPI
/*  Adjust for lack of user interaction when running with MPI. */
  INT Graph=0;                /* don't use graphical output with MPI */
#else
  INT Graph=1;                /* use graphical output */
#endif
int Profile=0;              /* Output timing for each device. */
int Append=0;               /* append to output file */
int Verbose=0;              /* verbose messages in output file */
int Mute=0;                 /* no default output to stdout */
int Optoff=0;               /* switch off options parsing, */
                            /* to allow for args that begin with `-' etc */
FILE *debug=0;              /* debug output file  */

/* others */
int narg=0;                 /* the additional arguments */
char **arg=NULL;            /* exported for def only !! */
int StopBatch=1;            /* return code for batch control */
char *WMODE[2]={"wt","at"}; /* write/append (text) modes */
int ndev=0;                 /* num of devices in current run */
Device dev[MAXDEV];         /* the array of devices - exported for init only!!*/
int idev;           	    /* device counter */

/* Static variables */
static char HEADLINE[256];	/* First line of output with package version variation and date */
static FILE *res;               /* log file */
static char logname[MAXPATH], debugname[MAXPATH];
static int iarg, argc_new;      /* argument counters */

/*  Timing variables */
static time_t start, finish;
static double run_start_time, run_end_time, device_start_time, device_end_time;

/* The Wallclock time, measured differently in different modes */
static double Wtime(void) {
  #if MPI
    return MPI_Wtime();
  #else
    return (double)clock()/(double)CLOCKS_PER_SEC;
  #endif
}

/* Generic message procedure, writes both to stdout and log file */
int MESSAGE(char *fmt, ...) {
  if (mpi_rank==0) { /*  Avoid repeated messages on MPI. */
    char s[512], *p;
    va_list argptr;
    va_start(argptr, fmt);
    vsprintf(s, fmt, argptr);                 /* form the message */
    va_end(argptr);
    p=s; if (s[0]=='\x01') p++;               /* \x01 - stdout mutes */
    if (res) {
          fputs(p,res);
          if(fflush_or_not==1)
          fflush(res);
          }                                   /* to log file - always */
    if (p>s) return 0;                        /* not to stdout */
    if (Mute==0) fputs(p,stdout); return(1);  /* to stdout */
  }
}

/* Same, but working from any thread:       */
/* useful when only one thread generates it */
int URGENT_MESSAGE(char *fmt, ...) {
  char s[512], *p;
  va_list argptr;
  va_start(argptr, fmt);
  vsprintf(s, fmt, argptr);                 /* form the message */
  va_end(argptr);
  p=s; if (s[0]=='\x01') p++;               /* \x01 - stdout mutes */
  if (res) {
    fputs(p,res);
    if(fflush_or_not==1)
      fflush(res);
  }                                   /* to log file - always */
  if (p>s) return 0;                        /* not to stdout */
  if (Mute==0) fputs(p,stdout); return(1);  /* to stdout */
}

int DEBUG(char *fmt, ...) {
  va_list argptr;
  if NOT(debug) return(1);
  va_start(argptr, fmt);
  vfprintf(debug, fmt, argptr);             /* form the message */
  va_end(argptr);
  /* if(fflush_or_not==1) */ /* debug is about thoroughness, not performance! */
  fflush(debug);
  return(1);
}

jmp_buf errjmp;
void ERROR(char *fmt, ...) {
  char s[4092], *p;
  va_list argptr;
  va_start(argptr, fmt);
  vsprintf(s, fmt, argptr);                     /* form the message */
  va_end(argptr);
  p=s; if (s[0]=='\x01') p++;                   /* \x01 - stdout mutes */
  if (res) {     /* to log file */
     fprintf(res,"\n%s",p);
     if(fflush_or_not==1) fflush(res);
  }
  if(mpi_rank == 0){
    if (p<=s) printf("\n%s",p); /* to stdout */
  }
  longjmp(errjmp,1);
}

/* Print a usage message. */
#define HELPMSG \
"Usage: \n"\
"   Beatbox [options] [--] <parameter_file> [<log_file> [additional_args..]]"\
"\nOptions: \n" \
"   -- : switches the parsing of further options off.\n"\
"   -append : appends to the output file.\n"\
"   -debug <one of stdout, stderr or a filename> : to output debugging info to.\n"\
"   -help : print this message to stdout and exit.\n"\
"   -log <filename> : to log to, else if no filename defaults to \"parameter_file.log\".\n"\
"   -mute : mutes output to stdout.\n"\
"   -nograph : do not produce graphical output (off for MPI-based jobs).\n"\
"   -profile : outputs timings for each device.\n"\
"   -verbose : produces verbose messages in the output file.\n"

#define HELP() ERROR("%s%s",HEADLINE,HELPMSG)

#ifdef UNIX
#include <unistd.h>
int    _argc; 
char **_argv;
extern char WindowName[], IconName[];
#endif

/**********************
 * Start main program *
 **********************/
int main (int argc, char **argv) {

  int id;		/* secondary device loops counter */
  int rc;		/* return code of a device */
  int *timesCalled;	/* number of times each device was executed */
  double spent;		/* time spent for a device call */
  double *totalSpent;	/* total time spent by each device */
  #if MPI
  int localrc;		/* return code of a device in this thread */
  double totalspent;	/* time spent by a device call in all threads */
  #endif

  sprintf(HEADLINE,
	  "#! %s %s version compiled %s %s\n"
	  "#------------------------------------------------------------------------------------\n"
	  ,VERSTRING,VARIATION,__DATE__,__TIME__);
	  /* "#! %s.%s %s version compiled %s %s\n" */
	  /* "#------------------------------------------------------------------------------------\n" */
	  /* ,VERSTRING,SUBVERSION,VARIATION,__DATE__,__TIME__); */

  #if MPI
    /*  Start the MPI processes. */
    mpi_errno = MPI_Init(NULL, NULL);
    CHECK_MPI_SUCCESS("Could not initialise MPI.")
      
    mpi_errno = MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank);
    CHECK_MPI_SUCCESS("Could not get my rank.")
  #else
    mpi_rank=0;
  #endif
        
  #ifdef UNIX /* Generate a tile for the X window */
    _argc=argc; _argv=argv;
    gethostname(IconName,1024); 
    strcat(IconName,":"); 
    strcat(IconName,argv[0]);
    gethostname(WindowName,1024); 
    strcat(WindowName,":");
    for(iarg=0;iarg<argc;iarg++) {
       strcat(WindowName,"  ");
       strcat(WindowName,argv[iarg]);
    }
  #endif

  strcpy(dev[idev=0].n,"Beatbox");
  if (setjmp(errjmp)) { StopBatch=1; goto Exit;}

  /*********************/
  /* Parse arguments   */
  /*********************/

  /* First, extract options */
  argc_new=1;
  for (iarg=1;iarg<argc;iarg++) {
    if (argv[iarg][0] && strchr("-",argv[iarg][0]) && !Optoff) {
      #define CASE(a) else if (0==stricmp(argv[iarg]+1,a))
      if (argv[iarg][1]=='?')   HELP();
      CASE("-")         Optoff   = 1;
      CASE("append")    Append   = 1;
      CASE("debug")     {
	iarg++;
	if (iarg>=argc)
	  ERROR("debug file name is not given after -debug option");
	if (0==strcmp(strcpy(debugname,argv[iarg]),"stdout"))
	  debug=stdout;
	else if (0==strcmp(debugname,"stderr"))
	  debug=stderr;
	else
	  debug=fopen(debugname,"wt");
      }
      CASE("help")      HELP();
      CASE("log")       {
	iarg++;
	if (iarg>=argc)
	  ERROR("log file name is not given after -log option");
	strcpy(logname, argv[iarg]);
      }
      CASE("mute")      Mute    = 1;
      CASE("nograph")   Graph   = 0;
      CASE("profile")   Profile = 1;
      CASE("verbose")   Verbose = 1;
      else              ERROR("Unknown option %s",argv[iarg]);
    } else {
      argv[argc_new++] = argv[iarg];
    }
  }
  argc = argc_new;

  /* Check num of arguments */
  if (argc<2 || argv[1][0]=='?') HELP();

  /* First arg - input file */
  if (!qopen(argv[1])) ERROR("Cannot open input file %s", argv[1]);

  /* log file */
  if NOT(*logname) fputext(argv[1],logname,"log");  /* default */
  if (0==strcmp(logname,null)) {
    res=NULL; 
    *logname='\0';
  } else if NOT(res=fopen(logname,WMODE[Append])) {
      ERROR("Cannot open log file %s", logname);
  }
  /* Next are additional parameters */
  if (argc>2) {narg = argc-2; arg = argv+2;}

  /**********************/
  /* Header of log file */
  /**********************/
  MESSAGE0(HEADLINE);
  time(&start);
  MESSAGE("execution begin at %24.24s $\n",ctime(&start));
  MESSAGE1("Input file %s ",argv[1]);
  if (narg) {
    MESSAGE0("with additional argument");
    if(narg>1) MESSAGE0("s");
    for (iarg=0;iarg<narg;iarg++)
      MESSAGE1(" \"%s\"",arg[iarg]);
    MESSAGE0(" $\n");
  } else {
    MESSAGE0("without additional arguments $\n");
  }

  MESSAGE0("with options:");
  if(Append) MESSAGE0(" append"); else MESSAGE0(" noappend");
  if(debug) MESSAGE1(" debug=%s",debugname); else MESSAGE0(" nodebug");
  if(Graph) MESSAGE0(" graph");else MESSAGE0(" nograph");
  if(Profile) MESSAGE0(" profile");else MESSAGE0(" noprofile");
  if(Verbose) MESSAGE0(" verbose");else MESSAGE0(" noverbose");
  if(*logname) MESSAGE1(" logname=%s",logname);else MESSAGE0(" logname=default");
  MESSAGE0("\n");
  
  /**************************************/
  /* Read input file and build the task */
  /**************************************/
  if(!init()) {
    strcpy(dev[idev=ndev].n,"init");
    ERROR("Failure in reading input file");
  }
  
  if (!ndev) {
    strcpy(dev[idev=ndev].n,"init");
    ERROR("No devices were created");
  }

  /********************/
  /* Perform the task */
  /********************/
  #if MPI
    /*  Inactive processes can wait for the others at the exit barrier. */
    if (I_AM_IDLE) goto Exit;
  #endif

  t = 0; /* Set the main loop counter to 0 */

  /*  Timing */
  run_start_time = Wtime();
  
  /*  Initialize profiling */
  if (Profile) {
    CALLOC(totalSpent,ndev, sizeof(double));
    CALLOC(timesCalled,ndev, sizeof(int));
    MESSAGE("============================ PROFILING DATA ============================\n");
    MESSAGE("t       ,");
    for (id=0; id<ndev; id++) {
      MESSAGE("%8s",dev[id].n);
      if (id<(ndev-1)) {
        MESSAGE(",");
      } else {
        MESSAGE("\n");
      }
    }
  }
  
  /*********************************************************/
  /* Time loop. Exit for loop when any Device returns a 0. */
  /*********************************************************/
  for (;;) { 
    DEBUG("t=%ld:",t);
    for (idev=0; idev<ndev; idev++) {  /* Loop over the active devices */
      #define d (dev[idev])
      if (Profile) {
        /*  Add timestep to profile. */
        if(idev==0){
          MESSAGE("%08d,",t);
        }
      }
      if (*(d.c)) { /* is the device condition met? */
	DEBUG(" %d(%s)",idev,d.n);
        if (Profile) device_start_time = Wtime();
	#if MPI
	  localrc=d.p(d.s, d.w, d.par, d.sync, d.alwaysRun);
          if (Profile) device_end_time = Wtime();
          MPI_Barrier(ALL_ACTIVE_PROCS);
	  mpi_errno = MPI_Reduce(&localrc, &rc, 1, MPI_INTEGER, MPI_BAND, 0, ALL_ACTIVE_PROCS);
	  CHECK_MPI_SUCCESS("Couldn't carry out reduction operation.")
	  mpi_errno = MPI_Bcast(&rc, 1, MPI_INTEGER, 0, ALL_ACTIVE_PROCS);
	  CHECK_MPI_SUCCESS("Couldn't carry out broadcast operation.")
	#else
          if (Graph && (d.w.row1>d.w.row0 || d.w.col1>d.w.col0)) {
            opengraph(); /* if X unavailable, proceed without */
            if (!d.w.drawn) {
               SetWindow(d.w);
               Frame(d.s); 
               d.w.drawn=1;
            }
          }
          /*
            The same structure is used in the serial and parallel codes.
            So the structure dev[idev].XXXX has some dangling components.
          */
          rc=d.p(d.s, d.w, d.par);
          if (Profile) device_end_time = Wtime();
  	#endif

	if (rc==0)
	  break;

        if (Profile) {
          spent = (device_end_time - device_start_time);
	  #if MPI
	  mpi_errno = MPI_Reduce(&spent, &totalspent, 1, MPI_DOUBLE, MPI_SUM, 0, ALL_ACTIVE_PROCS);
	  spent=totalspent/num_active_procs;
	  #endif
          DEBUG("%8f",spent);
          totalSpent[idev] += spent;
          timesCalled[idev] ++;
        }
      } else { /*  this device is not run now */
        if (Profile) DEBUG("--------");
      }
      
      /*  Separator */
      if (Profile) DEBUG((idev<(ndev-1))?",":"\n");

    } /* End the device loop */

    if (rc==0) {
      break; 
    }
    t++; /* increment the loop counter */
    DEBUG("\n");
 
   }; /* End main loop */

/*************************/
/* Terminate the program */
/*************************/
Exit: /* this the target for the long jump */

#if MPI
  /* the barrier has to be for all threads not just active ones */
  MPI_Barrier(MPI_COMM_WORLD);
  /* MESSAGE("\nall threads are through the exit barrier\n"); */

  if (!I_AM_IDLE) {
#endif
    /* more than 1 proc is hopefully active: run_end_time is for each proc. */
    run_end_time = Wtime();
    /* but it is reported only for rank = 0 */
    MESSAGE("\n\nTIMING INFO:\nThis run took %f seconds.\n\n", 
            run_end_time-run_start_time);
    
    if (Profile) {
      MESSAGE("Profiling summary:\n");
      MESSAGE("%-4s%-20s\t%8s\t%16s\t%16s\n","No","Device name","# calls","Time per step","Time per call");
      for(id=0;id<ndev;id++) {
	int tc=timesCalled[id];
	MESSAGE("%3d %-20s\t%8d\t%16.8f\t%16.8f\n",
		id,
		dev[id].n,
		timesCalled[id],
		totalSpent[id]/(double)t,
		totalSpent[id]/(tc?tc:1)
		);
      }	
      free(timesCalled);
      free(totalSpent);
    }

    if (graphon) {
      sleep(StopBatch==0?0.2:0.0);
      closegraph();
      graphon=0;
    }
    MESSAGE("\n%s finished at t=%ld by device %d \"%s\"\n",
            VERSTRING,t,idev,dev[idev].n);
    time(&finish);
    MESSAGE("%24.24s\n",ctime(&finish));
    MESSAGE("======================================================\n");
    if(res) fclose(res);
    if(debug) fclose(debug);

    /* Terminate the other devices cleanly */
    term();

#if MPI
  } /* If(!I_AM_IDLE) */

  /* Stop the MPI processes */
  MPI_Finalize();
#endif

  return(StopBatch);
} /* End of main */

