/**
 *  Copyright (C) (2010-2012) Vadim Biktashev, Alexander Karpov, 
 *                            Irina Biktasheva, Ross McFarlane;
 *  Copyright (C) 2012 Sanjay Kharche
 *
 *  This file is part of Beatbox.
 *
 *   Beatbox is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Beatbox is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Beatbox.  If not, see <http://www.gnu.org/licenses/>.
 */


/* Output a 4D box to a file, each mode value reduced to unsigned short */
/* File format: pure information, without markers 			*/
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include "system.h"
#include "beatbox.h"
#include "bikt.h"
#include "device.h"
#include "state.h"
#include "qpp.h"
#include "sequence.h"
#include "byte.h"
#include "mpi_io_choice.h"

typedef struct {
  int putsize;
  int append;
  sequence file;
  real u0,u1;			/* range of values to transfrom node to a byte */
  long int dx, dy, dz, dv;		/* to output decimated data */
#if MPI
  MPI_Datatype filetype;
  int bufsize;
  char *buf;
#endif
} STR;

RUN_HEAD(byteout)
  DEVICE_CONST(int,append) DEVICE_CONST(int,putsize)
  DEVICE_VAR(sequence,file)
  DEVICE_CONST(real,u0) DEVICE_CONST(real,u1)
  DEVICE_CONST(long int,dx) DEVICE_CONST(long int,dy) DEVICE_CONST(long int,dz) DEVICE_CONST(long int,dv) 
#if MPI
  DEVICE_CONST(MPI_Datatype,filetype)
  DEVICE_CONST(int,bufsize)
  DEVICE_ARRAY(char,buf)
  int ind = 0;
#endif
  int v, x, y, z, b;
  
  if (!file->f) return 1;
	
#if MPI
  /*  Set view for collective writing. Displace by length of header. */
  mpi_errno = MPI_File_set_view(file->f, 0, MPI_CHAR, filetype, "native", MPI_INFO_NULL);
  CHECK_MPI_SUCCESS("Couldn't set view for writing.")
#else
  if (putsize) {
    putc(s.v1-s.v0+1,file->f);putc(s.x1-s.x0+1,file->f);
    putc(s.y1-s.y0+1,file->f);putc(s.z1-s.z0+1,file->f);
  }
#endif
  for(z=s.z0;z<=s.z1;z+=dz) {
    for(y=s.y0;y<=s.y1;y+=dy) {
      for(x=s.x0;x<=s.x1;x+=dx) {
	for(v=s.v0;v<=s.v1;v+=dv) {
	  b = Byte(x,y,z,v,u0,u1);
#if MPI
	  buf[ind++] = b;
#else
	  putc(b,file->f);
#endif
	} /*  v */
      } /*  x */
    } /*  y */
  } /*  z */

#if MPI
  /*  Write buffer to file. */
  if(blocking_or_contiguous==0)
    mpi_errno = MPI_File_write(file->f, buf, bufsize, MPI_CHAR, MPI_STATUS_IGNORE);
  else
    mpi_errno = MPI_File_write_all(file->f, buf, bufsize, MPI_CHAR, MPI_STATUS_IGNORE);
  CHECK_MPI_SUCCESS("Couldn't write to file.")
#endif

  if (append){
#if MPI
#else
    if(fflush_or_not==1) fflush(file->f);
#endif
  } else {
    nextq(file);
  }/*  if(append) */
RUN_TAIL(byteout)

/*******************/
DESTROY_HEAD(byteout)
#if MPI
        if(s.runHere) free(S->buf);
#else
	if (S->file.f) fclose(S->file.f); S->file.f=NULL;
#endif
DESTROY_TAIL(byteout)

/********************/
CREATE_HEAD(byteout) {
  ACCEPTI(putsize,0,0,1);
  ACCEPTI(append,0,0,1);
  ACCEPTQ(file,S->append?"at":"wt",NULL);
  ACCEPTR(u0,RNONE,RNONE,RNONE);
  ACCEPTR(u1,RNONE,RNONE,RNONE);
  ASSERT(S->u1!=S->u0);
  ACCEPTL(dx,1,1,xmax);
  ACCEPTL(dy,1,1,ymax);
  ACCEPTL(dz,1,1,zmax);
  ACCEPTL(dv,1,1,vmax);

#if MPI
  if(S->putsize){
    MESSAGE("WARNING: putsize is not available under MPI. Byteout will carry on regardless.\n");
  }
  if(dev->s.runHere){
#define NDIMS 4
    MPI_Datatype filetype;
    MPI_Datatype sourceType;
    int sizes[NDIMS],subsizes[NDIMS],starts[NDIMS];
    int bufsize;
    
    /*  Global space dimensions. */
    int xsize = (dev->s.global_x1 - dev->s.global_x0) + 1;
    int ysize = (dev->s.global_y1 - dev->s.global_y0) + 1;
    int zsize = (dev->s.global_z1 - dev->s.global_z0) + 1;
    int vsize = (dev->s.v1 - dev->s.v0) + 1;
    
    /*  Local space dimensions */
    int local_xsize = (dev->s.x1 - dev->s.x0) + 1;
    int local_ysize = (dev->s.y1 - dev->s.y0) + 1;
    int local_zsize = (dev->s.z1 - dev->s.z0) + 1;
    int local_vsize = vsize;
    
    /*  Dimensions of the global space. */
    sizes[0] = xsize;
    sizes[1] = ysize;
    sizes[2] = zsize;
    sizes[3] = vsize;

    /*  Dimensions of the local space. */
    subsizes[0] = local_xsize;
    subsizes[1] = local_ysize;
    subsizes[2] = local_zsize;
    subsizes[3] = local_vsize;
    
    starts[0] = dev->s.x0 - dev->s.global_x0;
    starts[1] = dev->s.y0 - dev->s.global_y0;
    starts[2] = dev->s.z0 - dev->s.global_z0;
    starts[3] = 0;

    mpi_errno = MPI_Type_create_subarray(NDIMS, sizes, subsizes, starts, MPI_ORDER_C, MPI_CHAR, &filetype);
    CHECK_MPI_SUCCESS("Couldn't define filetype.")
      mpi_errno = MPI_Type_commit(&filetype);
    CHECK_MPI_SUCCESS("Couldn't commit filetype.")
      
      bufsize = local_xsize*local_ysize*local_zsize*local_vsize;
    CALLOC(S->buf,bufsize,sizeof(char));
    
    S->filetype = filetype;
    S->bufsize = bufsize;
    
#undef NDIMS
  }/*  if runHere */
#endif

} CREATE_TAIL(byteout,0)
