/**
 *  Copyright (C) (2010-2012) Vadim Biktashev, Alexander Karpov, 
 *                            Irina Biktasheva, Ross McFarlane;
 *  Copyright (C) 2012 Sanjay Kharche
 *
 *  This file is part of Beatbox.
 *
 *   Beatbox is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Beatbox is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Beatbox.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <mpi.h>
#include <stdio.h>
#include<math.h>
#include "error.h"
#include "system.h"
#include "beatbox.h"
#include "device.h"
#include "state.h"
#define OWN
#include "decomp.h"
#undef OWN

#define SK_DECOM 1 // 1 if using SK code for factoring the number of processes.

/* 8< decomp_getworstsubdomaindimensions */
void decomp_getWorstSubdomainDimensions(Decomposition d, int *xlen, int *ylen, int *zlen){
/*
	*xlen = (xmax / d.nx) + (xmax % d.nx == 0)?0:1;
	*ylen = (ymax / d.ny) + (ymax % d.ny == 0)?0:1;
	*zlen = (zmax / d.nz) + (zmax % d.nz == 0)?0:1;
*/

	*xlen = (xmax / d.nx) + (xmax % d.nx);
	*ylen = (ymax / d.ny) + (ymax % d.ny);
	*zlen = (zmax / d.nz) + (zmax % d.nz);


}
/* >8 decomp_getworstsubdomaindimensions */

/* 8< decomp_getworstsubdomainvolume */
int decomp_getWorstSubdomainVolume(Decomposition d){
	int xlen, ylen, zlen;
	decomp_getWorstSubdomainDimensions(d, &xlen, &ylen, &zlen);
	return xlen * ylen * zlen;
}
/* >8 decomp_getworstsubdomainvolume */

/* 8< decomp_getworstsubdomainsurfacearea */
int decomp_getWorstSubdomainSurfaceArea(Decomposition d){
	int xlen, ylen, zlen;
	decomp_getWorstSubdomainDimensions(d, &xlen, &ylen, &zlen);
	return 2*((xlen*ylen)+(ylen*zlen)+(xlen*zlen));
}
/* >8 decomp_getworstsubdomainsurfacearea */

/* 8< decomp_getmetric */
float decomp_getMetric(Decomposition d){
	return (float)	(decomp_getWorstSubdomainVolume(d) * DECOMP_VOLUME_ADJUST) +
					(decomp_getWorstSubdomainSurfaceArea(d) * DECOMP_SURFACE_AREA_ADJUST);
}
/* >8 decomp_getmetric */

/* 8< decomp_getbest */
Decomposition decomp_getBest(Decomposition a, Decomposition b){	
	return (decomp_getMetric(a) < decomp_getMetric(b)) ? a : b;
}
/* >8 decomp_getbest */

/* 8< decompose */
int decompose(int num_procs, int *nx, int *ny, int *nz){	
	int nx_max,ny_max,nz_max;
	int nx_tmp_max,ny_tmp_max,nz_tmp_max;
	Decomposition best_yet, newcomer;
	
/*
Factorization variables.
*/

int no,x;
int factors[10000],max_factors = 0,factor_counter = 0;



	/*	Maxima for nx, ny and nz
	 *	At most one processor along an inactive axis.
	 *	No more than one processor per point.
	 */
	nx_max = (ONE) ? min(xmax, num_procs) : 1;
	ny_max = (TWO) ? min(ymax, num_procs) : 1;
	nz_max = (TRI) ? min(zmax, num_procs) : 1;
	
	best_yet.nx = 1;
	best_yet.ny = 1;
	best_yet.nz = 1;
	
/*
For any number of nodes, the decomp_getBest returns only 1.
OK. So by hard coding this one, I can see that the run time is reduced.
If I set best_yet.nx to be 4, then the run time is reduced from 
*/
	for(newcomer.nx = 1; newcomer.nx <= nx_max; newcomer.nx++){
		ny_tmp_max = min(ny_max, num_procs/newcomer.nx);
		for(newcomer.ny = 1; newcomer.ny <= ny_tmp_max; newcomer.ny++){
			nz_tmp_max = min(nz_max, num_procs/(newcomer.nx*newcomer.ny));
			for(newcomer.nz = 1; newcomer.nz <= nz_tmp_max; newcomer.nz++){
				best_yet = decomp_getBest(newcomer, best_yet); // this simply returns a 1 each time
			}
		}
	}
	

if(SK_DECOM==1){

printf("Executing simple factoring of processes \n");

/* TEST: Hard code the 4 threads to see if it helps any. Answer: it does help! */
/*
	best_yet.nx = 4;
	best_yet.ny = 1;
	best_yet.nz = 1;
*/

/* this is 1D case specific. */
/*
The conditions to check for this are:
1. get the num_procs: already done.
2. No geometry should be on.
3. It should be 1D along the x axis
4. Always in the MPI #if: this is already true by the time the code gets here.

Some more conditions have to be considered. The way the num_procs are
being partitioned, it assumes that num_procs is at least size
dim. This need not always be true.
*/

// printf("%d %d %d %d %d\n",ONE,TWO,TRI,dim,GEOMETRY_ON);

if(dim==1){ // this condition alone makes sure that we have satisfied conditions 1-4.
	best_yet.nx = num_procs;
	best_yet.ny = 1;
	best_yet.nz = 1;
} // end of dim == 1

if(dim==2 /* &&!GEOMETRY_ON */ ){ // at this stage, there could be a slice. so dont include geometry yet.

if(num_procs==1||num_procs==2){
	best_yet.nx = num_procs;
	best_yet.ny = 1;
	best_yet.nz = 1;
}
else
{

no = num_procs;

for(x=1; x<=no; x++)
if(no%x==0){
	factors[factor_counter] = x;
	factor_counter++;
} // end of if

// if it is not a perfect square, then the number of factors is even by now.
if(factor_counter%2==0){
//	printf("%d %d \n",factors[(factor_counter-1)/2],factors[(factor_counter+1)/2]);
	best_yet.nx = factors[(factor_counter-1)/2];
	best_yet.ny = factors[(factor_counter+1)/2];
	best_yet.nz = 1;
}

// a perfect square.
if(factor_counter%2!=0){
//	printf("%d %d \n",factors[(factor_counter-1)/2],factors[(factor_counter-1)/2]);
	best_yet.nx = factors[(factor_counter-1)/2];
	best_yet.ny = factors[(factor_counter-1)/2];
	best_yet.nz = 1;
}


} // end of the else statement

} // end of dim==2

if(dim==3 /*&&!GEOMETRY_ON */){ // I have not done the 3D partition correctly yet. The 3D geometry also works. This code is geared towards 1,2,3, and perfect cube num_procs. The code for other values of num_procs is required to cover all bases.

if(num_procs==3||num_procs==2||num_procs==1){
	best_yet.nx = num_procs;
	best_yet.ny = 1;
	best_yet.nz = 1;
}
else
if(num_procs==2*2*2||num_procs==3*3*3||num_procs==4*4*4||num_procs==5*5*5||num_procs==6*6*6||num_procs==7*7*7||num_procs==8*8*8||num_procs==9*9*9||num_procs==10*10*10||num_procs==11*11*11) // this should be ok for now.
{ // this is only true for full cubes as of now. Improve the cube root finding, or hard code it in for the first several cubes.
	best_yet.nx = (int)ceil(pow((double)num_procs,1.0/3.0));
	best_yet.ny = (int)ceil(pow((double)num_procs,1.0/3.0));
	best_yet.nz = (int)ceil(pow((double)num_procs,1.0/3.0));
}
else
{

no = num_procs;

for(x=1; x<=no; x++)
if(no%x==0){
	factors[factor_counter] = x;
	factor_counter++;
} // end of if

// if it is not a perfect square, then the number of factors is even by now.
if(factor_counter%2==0){
//	printf("%d %d \n",factors[(factor_counter-1)/2],factors[(factor_counter+1)/2]);
	best_yet.nx = factors[(factor_counter-1)/2];
	best_yet.ny = factors[(factor_counter+1)/2];
	best_yet.nz = 1;
}

// a perfect square.
if(factor_counter%2!=0){
//	printf("%d %d \n",factors[(factor_counter-1)/2],factors[(factor_counter-1)/2]);
	best_yet.nx = factors[(factor_counter-1)/2];
	best_yet.ny = factors[(factor_counter-1)/2];
	best_yet.nz = 1;
}


} // end of if else

} // end of dim==3

} // end of SK_DECOMP

	*nx = best_yet.nx;
	*ny = best_yet.ny;
	*nz = best_yet.nz;

	return best_yet.nx * best_yet.ny * best_yet.nz;
}
/* >8 decompose */

/* 8< defining_all_active_procs */
int decomp_globalCommunicator(int old_rank, MPI_Comm *new_comm, int *new_rank){
	int i,is_idle;
	MPI_Group old_group, new_group;
	if(old_rank >= num_active_procs){
		is_idle = 1;
	}else{
		is_idle = 0;
	}
	int group_ranks[num_active_procs];
	/* Ranks 0 --> num_active_procs-1 are included in new communicator. */
	for(i=0; i<num_active_procs; i++){
		group_ranks[i] = i;
	}

	mpi_errno = MPI_Comm_group(MPI_COMM_WORLD, &old_group);
	CHECK_MPI_SUCCESS("Couldn't get old communicator's group.")

	mpi_errno = MPI_Group_incl(old_group, num_active_procs, group_ranks, &new_group);
	CHECK_MPI_SUCCESS("Couldn't make new group.")

	mpi_errno = MPI_Comm_create(MPI_COMM_WORLD, new_group, new_comm);
	CHECK_MPI_SUCCESS("Couldn't create new communicator.")

	if(!is_idle){
		/*	Redefine rank based on new communicator
		 *	This shouldn't have changed, but we can't guarantee it.	*/
		mpi_errno = MPI_Comm_rank(*new_comm, new_rank);
		CHECK_MPI_SUCCESS("Couldn't get my new rank.")
	}else{
		/* If the process is idle, we keep the old rank so that it can report. */
		*new_rank = old_rank;
	}
	
	return is_idle;
}
/* >8 defining_all_active_procs */	

/* 8< compute_superindices */
void decomp_getSuperindices(int rank, int *ix, int *iy, int *iz){
	int rt;
	*ix = rank / (mpi_ny * mpi_nz);
	rt = rank % (mpi_ny * mpi_nz);
	*iy = rt / mpi_nz;
	*iz = rt % mpi_nz;
}
/* >8 compute_superindices */

int decomp_getSubdomainDimensions(){
	/*	Compute bounds of all subdomains:
	 *	Each process contains active points from local_*min to local_*max -1	*/
	CALLOC(subdomain_limits,num_active_procs,sizeof(Subdomain));
	int rank, x_slice_size, y_slice_size, z_slice_size, x_remainder, y_remainder, z_remainder;
	int ix, iy, iz;
	int i;

/* 8< compute_subdomain_bounds */	
	/*	Approximate length of subdomain along each axis.	*/
	x_slice_size = (xmax-(ONE*2)) / mpi_nx;
	y_slice_size = (ymax-(TWO*2)) / mpi_ny;
	z_slice_size = (zmax-(TRI*2)) / mpi_nz;
	
	/* Remainders from i[xyz] * [xyz]_slice_size */
	/* The remainder will be shared amongst the first processes on the axis */
	x_remainder  = (xmax-(ONE*2)) % mpi_nx;
	y_remainder  = (ymax-(TWO*2)) % mpi_ny;
	z_remainder  = (zmax-(TRI*2)) % mpi_nz;
	
	for(rank=0;rank<num_active_procs;rank++){
		decomp_getSuperindices(rank,&ix,&iy,&iz);

		/*Lower x bound	*/
		subdomain_limits[rank].local_xmin	= ONE + (ix * x_slice_size) + (ix < x_remainder? ix :x_remainder);
		/*Upper x bound	*/
		subdomain_limits[rank].local_xmax	= ONE + ((ix+1) * x_slice_size) + (ix < x_remainder? (ix+1) :x_remainder);
		
		/*Lower y bound	*/
		subdomain_limits[rank].local_ymin	= TWO + (iy * y_slice_size) + (iy < y_remainder? iy :y_remainder);
		/*Upper y bound	*/
		subdomain_limits[rank].local_ymax	= TWO + ((iy+1) * y_slice_size) + (iy < y_remainder? (iy+1) :y_remainder);
			
		/*Lower z bound	*/
		subdomain_limits[rank].local_zmin	= TRI + (iz * z_slice_size) + (iz < z_remainder? iz :z_remainder);
		/*Upper z bound	*/
		subdomain_limits[rank].local_zmax	= TRI + ((iz+1) * z_slice_size) + (iz < z_remainder? (iz+1) :z_remainder);
	} /* for(rank) */
	
	/*	Assign own subdomain boundaries */
	local_xmin = subdomain_limits[mpi_rank].local_xmin;
	local_xmax = subdomain_limits[mpi_rank].local_xmax;
	local_ymin = subdomain_limits[mpi_rank].local_ymin;
	local_ymax = subdomain_limits[mpi_rank].local_ymax;
	local_zmin = subdomain_limits[mpi_rank].local_zmin;
	local_zmax = subdomain_limits[mpi_rank].local_zmax;
/* >8 compute_subdomain_bounds */
	
	return 1;
}

void decomp_getSubdomainSize(){
	xlen = (local_xmax-local_xmin)+(ONE*2);
	ylen = (local_ymax-local_ymin)+(TWO*2);
	zlen = (local_zmax-local_zmin)+(TRI*2);
}

/* 8< mpi_decomp_define_halo_types */
void decomp_defineHaloTypes(){
	/*	Types for Halo Swapping
	 *
	 *	Labelled as XN, XP, YN, YP, ZN, ZP
	 *	where, e.g. XN is the YZV hyperplane at local_xmin.
	 *	ZP is the XYV hyperplane at local_zmax.
	 *
	 *	ZP_Halo_Type is the halo at local_xmax.
	 *	YN_Halo_Type is the halo at local_ymin-1.
	 *
	 *	Hyperplanes exchanged in the y axis are widened in
	 *	the x axis, and those exchanged in the z axis are widened in
	 *	in the x and y axes. This includes points from diagonally 
	 *	neighbouring subdomains, which prevents the need for exchanging 
	 *	corner points explicitly between these processes.
	 *	-----------------------------------	*/
	#define NDIMS 4
	int sizes[NDIMS],subsizes[NDIMS],starts[NDIMS];
	sizes[0] = xlen;
	sizes[1] = ylen;
	sizes[2] = zlen;
	sizes[3] = vmax;

	if(mpi_nx > 1){
		/*	XN	*/
		subsizes[0] = 1;
		subsizes[1] = local_ymax-local_ymin;
		subsizes[2] = local_zmax-local_zmin;
		subsizes[3] = vmax;
	
		starts[0] = ONE;
		starts[1] = TWO;
		starts[2] = TRI;
		starts[3] = 0;
		
		mpi_errno = MPI_Type_create_subarray(NDIMS, sizes, subsizes, starts, MPI_ORDER_C, MPI_DOUBLE, &XN_Type);
		CHECK_MPI_SUCCESS("Couldn't define XN_Type.")
		mpi_errno = MPI_Type_commit(&XN_Type);
		CHECK_MPI_SUCCESS("Couldn't commit XN_Type.")

		/*	XN Halo	*/
		starts[0] -= 1;
		mpi_errno = MPI_Type_create_subarray(NDIMS, sizes, subsizes, starts, MPI_ORDER_C, MPI_DOUBLE, &XN_Halo_Type);
		CHECK_MPI_SUCCESS("Couldn't define XN_Halo_Type.")
		mpi_errno = MPI_Type_commit(&XN_Halo_Type);
		CHECK_MPI_SUCCESS("Couldn't commit XN_Halo_Type.")

		/*	XP	*/
		starts[0] = (local_xmax-1 + ONE) - local_xmin;
		starts[1] = TWO;
		starts[2] = TRI;
		starts[3] = 0;
		mpi_errno = MPI_Type_create_subarray(NDIMS, sizes, subsizes, starts, MPI_ORDER_C, MPI_DOUBLE, &XP_Type);
		CHECK_MPI_SUCCESS("Couldn't define XP_Type.")
		mpi_errno = MPI_Type_commit(&XP_Type);
		CHECK_MPI_SUCCESS("Couldn't commit XP_Type.")

		/*	XP Halo	*/
		starts[0] += 1;
		mpi_errno = MPI_Type_create_subarray(NDIMS, sizes, subsizes, starts, MPI_ORDER_C, MPI_DOUBLE, &XP_Halo_Type);
		CHECK_MPI_SUCCESS("Couldn't define XP_Halo_Type.")
		mpi_errno = MPI_Type_commit(&XP_Halo_Type);
		CHECK_MPI_SUCCESS("Couldn't commit XP_Halo_Type.")
	} /*	if(mpi_nx > 1)	*/
	if(mpi_ny > 1){
		/*	YN	*/
		subsizes[0] = xlen; /*	Widened to include magic corners.	*/
		subsizes[1] = 1;
		subsizes[2] = local_zmax-local_zmin;
		subsizes[3] = vmax;

		starts[0] = 0;		/*	0 to include magic corner.	*/
		starts[1] = TWO;
		starts[2] = TRI;
		starts[3] = 0;
		mpi_errno = MPI_Type_create_subarray(NDIMS, sizes, subsizes, starts, MPI_ORDER_C, MPI_DOUBLE, &YN_Type);
		CHECK_MPI_SUCCESS("Couldn't define YN_Type.")
		mpi_errno = MPI_Type_commit(&YN_Type);
		CHECK_MPI_SUCCESS("Couldn't commit YN_Type.")
	
		/*	YN Halo		*/
		starts[1] -= 1;
		mpi_errno = MPI_Type_create_subarray(NDIMS, sizes, subsizes, starts, MPI_ORDER_C, MPI_DOUBLE, &YN_Halo_Type);
		CHECK_MPI_SUCCESS("Couldn't define YN_Halo_Type.")
		mpi_errno = MPI_Type_commit(&YN_Halo_Type);
		CHECK_MPI_SUCCESS("Couldn't commit YN_Halo_Type.")

		/*	YP	*/
		starts[0] = 0;		/*	0 to include magic corner.	*/
		starts[1] = (local_ymax-1 + TWO) - local_ymin;
		starts[2] = TRI;
		starts[3] = 0;
		mpi_errno = MPI_Type_create_subarray(NDIMS, sizes, subsizes, starts, MPI_ORDER_C, MPI_DOUBLE, &YP_Type);
		CHECK_MPI_SUCCESS("Couldn't define YP_Type.")
		mpi_errno = MPI_Type_commit(&YP_Type);
		CHECK_MPI_SUCCESS("Couldn't commit YP_Type.")

		/*	YP Halo	*/
		starts[1] += 1;
		mpi_errno = MPI_Type_create_subarray(NDIMS, sizes, subsizes, starts, MPI_ORDER_C, MPI_DOUBLE, &YP_Halo_Type);
		CHECK_MPI_SUCCESS("Couldn't define YP_Type.")
		mpi_errno = MPI_Type_commit(&YP_Halo_Type);
		CHECK_MPI_SUCCESS("Couldn't commit YP_Type.")
	} /*	if(mpi_ny > 1)	*/
	if(mpi_nz > 1){
		/*	ZN	*/
		subsizes[0] = xlen;	/*	Widened to include magic corners	*/
		subsizes[1] = ylen;	/*	Widened to include magic corners	*/
		subsizes[2] = 1;
		subsizes[3] = vmax;

		starts[0] = 0;		/*	0 to include magic corner.	*/
		starts[1] = 0;		/*	0 to include magic corner.	*/
		starts[2] = TRI;
		starts[3] = 0;
		mpi_errno = MPI_Type_create_subarray(NDIMS, sizes, subsizes, starts, MPI_ORDER_C, MPI_DOUBLE, &ZN_Type);
		CHECK_MPI_SUCCESS("Couldn't define ZN_Type.")
		mpi_errno = MPI_Type_commit(&ZN_Type);
		CHECK_MPI_SUCCESS("Couldn't commit ZN_Type.")
	
		/*	ZN Halo	*/
		starts[2] -= 1;
		mpi_errno = MPI_Type_create_subarray(NDIMS, sizes, subsizes, starts, MPI_ORDER_C, MPI_DOUBLE, &ZN_Halo_Type);
		CHECK_MPI_SUCCESS("Couldn't define ZN_Halo_Type.")
		mpi_errno = MPI_Type_commit(&ZN_Halo_Type);
		CHECK_MPI_SUCCESS("Couldn't commit ZN_Halo_Type.")

		/*	ZP	*/
		starts[0] = 0;		/*	0 to include magic corner.	*/
		starts[1] = 0;		/*	0 to include magic corner.	*/
		starts[2] = (local_zmax-1 + TRI) - local_zmin;
		starts[3] = 0;
		mpi_errno = MPI_Type_create_subarray(NDIMS, sizes, subsizes, starts, MPI_ORDER_C, MPI_DOUBLE, &ZP_Type);
		CHECK_MPI_SUCCESS("Couldn't define ZP_Type.")
		mpi_errno = MPI_Type_commit(&ZP_Type);
		CHECK_MPI_SUCCESS("Couldn't commit ZP_Type.")

		/*	ZP Halo	*/
		starts[2] += 1;
		mpi_errno = MPI_Type_create_subarray(NDIMS, sizes, subsizes, starts, MPI_ORDER_C, MPI_DOUBLE, &ZP_Halo_Type);
		CHECK_MPI_SUCCESS("Couldn't define ZP_Halo_Type.")
		mpi_errno = MPI_Type_commit(&ZP_Halo_Type);
		CHECK_MPI_SUCCESS("Couldn't commit ZP_Halo_Type.")
	} /*	if(mpi_nz > 1)	*/
	
	#undef NDIMS
}

/* >8 mpi_decomp_define_halo_types */
