/**
 *  Copyright (C) (2010-2012) Vadim Biktashev, Alexander Karpov, 
 *                            Irina Biktasheva, Ross McFarlane;
 *  Copyright (C) 2012 Sanjay Kharche
 *
 *  This file is part of Beatbox.
 *
 *   Beatbox is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Beatbox is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Beatbox.  If not, see <http://www.gnu.org/licenses/>.
 */


/*  Utility functions for devices. */
#include <stdio.h>
#include <stdarg.h>
#include "dynamic.h"
#include "beatbox.h"
#include "state.h"
#define OWN
#include "device.h"
#undef OWN

#if MPI
/* 8< mpi_device_communicator */
/* 	Define new group and communicator for I/O, 
 *	including only instances with runHere = true.	*/
int deviceCommunicatorWithFirstRank(int runHere, MPI_Comm *new_comm, int *first){	
	/*  Group of existing communicator (ALL_ACTIVE_PROCS). */
	MPI_Group old_group;
	/*  Group for new communicator */
	MPI_Group new_group;
	/*  runHere flag from every process. */
	int runHeres[num_active_procs];
	/*  Number of processes in new group. */
	int group_count = 0;
	/*  Handy integers. */
	int i, next, success;
	
	mpi_errno = MPI_Allgather(&runHere, 1, MPI_INT, &runHeres, 1, MPI_INT, ALL_ACTIVE_PROCS);
	CHECK_MPI_SUCCESS("Couldn't gather runHere flags.")
	
	/*  Count running instances. */
	for(i=0; i<num_active_procs; i++){
		if(runHeres[i]){
			group_count++;
		}
	}
	
	/*  Identify ranks of running instances to include in new group. */
	int group_ranks[group_count];
	next = 0;
	for(i=0; i<num_active_procs; i++){
		if(runHeres[i]){
			group_ranks[next++] = i;
		}
	}
	
	*first = group_ranks[0];
	
	mpi_errno = MPI_Comm_group(ALL_ACTIVE_PROCS, &old_group);
	CHECK_MPI_SUCCESS("Couldn't get old communicator's group.")
	
	mpi_errno = MPI_Group_incl(old_group, group_count, group_ranks, &new_group);
	CHECK_MPI_SUCCESS("Couldn't make new group.")
	
	mpi_errno = MPI_Comm_create(ALL_ACTIVE_PROCS, new_group, new_comm);
	CHECK_MPI_SUCCESS("Couldn't create new communicator.")
	
	return 1;
}

/*  Overloaded for the majority of occasions where first isn't necessary. */
int deviceCommunicator(int runHere, MPI_Comm *new_comm){
	int first;
	return deviceCommunicatorWithFirstRank(runHere, new_comm, &first);
}
/* >8 mpi_device_communicator */

/* 8< mpi_haloswap */
int haloSwap(){
/* X AXIS */
	if(mpi_ix % 2 == 0){
		if(mpi_ix < mpi_nx-1){ 
			mpi_errno = MPI_Sendrecv(New, 1, XP_Type, XP_NEIGHBOUR, 0, New, 1, XP_Halo_Type, XP_NEIGHBOUR, 0, ALL_ACTIVE_PROCS, MPI_STATUS_IGNORE); 
		} 
	}else{ 
		/*  No need to check if ix > 0, since ix is odd here. */
			mpi_errno = MPI_Sendrecv(New, 1, XN_Type, XN_NEIGHBOUR, 0, New, 1, XN_Halo_Type, XN_NEIGHBOUR, 0, ALL_ACTIVE_PROCS, MPI_STATUS_IGNORE); 
	}
	CHECK_MPI_SUCCESS("Failed to swap buffers.")
	if(mpi_ix % 2 == 1){ 
		if(mpi_ix < mpi_nx-1){ 
			mpi_errno = MPI_Sendrecv(New, 1, XP_Type, XP_NEIGHBOUR, 0, New, 1, XP_Halo_Type, XP_NEIGHBOUR, 0, ALL_ACTIVE_PROCS, MPI_STATUS_IGNORE); 
		} 
	}else{ 
		if(mpi_ix > 0){ 
			mpi_errno = MPI_Sendrecv(New, 1, XN_Type, XN_NEIGHBOUR, 0, New, 1, XN_Halo_Type, XN_NEIGHBOUR, 0, ALL_ACTIVE_PROCS, MPI_STATUS_IGNORE); 
		} 
	}
	CHECK_MPI_SUCCESS("Failed to swap buffers.")

/* Y AXIS */ 
	if(mpi_iy % 2 == 0){ 
		if(mpi_iy < mpi_ny-1){ 
			mpi_errno = MPI_Sendrecv(New, 1, YP_Type, YP_NEIGHBOUR, 0, New, 1, YP_Halo_Type, YP_NEIGHBOUR, 0, ALL_ACTIVE_PROCS, MPI_STATUS_IGNORE);
		} 
	}else{ 
		if(mpi_iy > 0){ 
			mpi_errno = MPI_Sendrecv(New, 1, YN_Type, YN_NEIGHBOUR, 0, New, 1, YN_Halo_Type, YN_NEIGHBOUR, 0, ALL_ACTIVE_PROCS, MPI_STATUS_IGNORE); 
		}
	} 
	CHECK_MPI_SUCCESS("Failed to swap buffers.")
	if(mpi_iy % 2 == 1){ 
		if(mpi_iy < mpi_ny-1){ 
			mpi_errno = MPI_Sendrecv(New, 1, YP_Type, YP_NEIGHBOUR, 0, New, 1, YP_Halo_Type, YP_NEIGHBOUR, 0, ALL_ACTIVE_PROCS, MPI_STATUS_IGNORE); 
		} 
	}else{ 
		if(mpi_iy > 0){ 
			mpi_errno = MPI_Sendrecv(New, 1, YN_Type, YN_NEIGHBOUR, 0, New, 1, YN_Halo_Type, YN_NEIGHBOUR, 0, ALL_ACTIVE_PROCS, MPI_STATUS_IGNORE); 
		} 
	}
	CHECK_MPI_SUCCESS("Failed to swap buffers.") 

/* Z AXIS */
	if(mpi_iz % 2 == 0){ 
		if(mpi_iz < mpi_nz-1){ 
			mpi_errno = MPI_Sendrecv(New, 1, ZP_Type, ZP_NEIGHBOUR, 0, New, 1, ZP_Halo_Type, ZP_NEIGHBOUR, 0, ALL_ACTIVE_PROCS, MPI_STATUS_IGNORE); 
		} 
	}else{ 
		if(mpi_iz > 0){ 
			mpi_errno = MPI_Sendrecv(New, 1, ZN_Type, ZN_NEIGHBOUR, 0, New, 1, ZN_Halo_Type, ZN_NEIGHBOUR, 0, ALL_ACTIVE_PROCS, MPI_STATUS_IGNORE); 
		} 
	} 
	CHECK_MPI_SUCCESS("Failed to swap buffers.") 
	if(mpi_iz % 2 == 1){ 
		if(mpi_iz < mpi_nz-1){ 
			mpi_errno = MPI_Sendrecv(New, 1, ZP_Type, ZP_NEIGHBOUR, 0, New, 1, ZP_Halo_Type, ZP_NEIGHBOUR, 0, ALL_ACTIVE_PROCS, MPI_STATUS_IGNORE); 
		} 
	}else{ 
		if(mpi_iz > 0){ 
			mpi_errno = MPI_Sendrecv(New, 1, ZN_Type, ZN_NEIGHBOUR, 0, New, 1, ZN_Halo_Type, ZN_NEIGHBOUR, 0, ALL_ACTIVE_PROCS, MPI_STATUS_IGNORE); 
		} 
	}
	CHECK_MPI_SUCCESS("Failed to swap buffers.") 
	return 1;
}
/* >8 mpi_haloswap */
#endif

