/**
 *  Copyright (C) (2010-2012) Vadim Biktashev, Alexander Karpov, 
 *                            Irina Biktasheva, Ross McFarlane;
 *  Copyright (C) 2012 Sanjay Kharche
 *
 *  This file is part of Beatbox.
 *
 *   Beatbox is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Beatbox is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Beatbox.  If not, see <http://www.gnu.org/licenses/>.
 */


/* 
 * Dump contents of a 4D subset to a binary file
 */

#include <stdlib.h>
#include <stdio.h>

#include <stdarg.h>
#include "system.h"
#include "beatbox.h"
#include "bikt.h"
#include "state.h"
#include "device.h"
#include "qpp.h"
#include "mpi_io_choice.h"


typedef struct {
	char filename[MAXPATH];
#if MPI	
	MPI_File file; 				/*  File handle for collective I/O. */
	MPI_Datatype sourceType;	/*  Datatype describing subarray of New to be written. */
#else
	FILE *file; /*  File handle for sequential I/O. */
#endif
	int append;
} STR;

/*  ----------------------------------------------------------------// */

// /*#include argsused.h*/
/* 8< mpi_dump_run */
RUN_HEAD(dump)
int x, y, z, v; /*  Loop indices */

#if MPI /*  MPI version */
DEVICE_CONST(int, append)
DEVICE_CONST(MPI_File, file)
DEVICE_CONST(MPI_Datatype, sourceType)

/*  Rewind file if not appending */
if(!append) MPI_File_seek(file,0,MPI_SEEK_SET);

/*  Write to file. */
if(blocking_or_contiguous==0)
mpi_errno = MPI_File_write(file, New, 1, sourceType, MPI_STATUS_IGNORE);
else
mpi_errno = MPI_File_write_all(file, New, 1, sourceType, MPI_STATUS_IGNORE);

CHECK_MPI_SUCCESS("Couldn't write to file.")

#else /*  Sequential Version */
DEVICE_CONST(FILE *,file)
DEVICE_CONST(int, append)
if (!file) return 1;
if (!append) rewind(file);
if (s.x0==0 && s.x1==xmax-1 && s.y0==0 && s.y1==ymax-1 && s.z0==0 && s.z1==zmax-1 && s.v0==0 && s.v1==vmax-1){
	fwrite(New,sizeof(real),xmax*ymax*zmax*vmax,file);
}else{
	for(x=s.x0;x<=s.x1;x++){
		for(y=s.y0;y<=s.y1;y++){
			for(z=s.z0;z<=s.z1;z++){
				for(v=s.v0;v<=s.v1;v++){
					fwrite(New+ind(x,y,z,v),sizeof(real),1,file);
				}
			}
		}
	}
}

if(fflush_or_not==1)
fflush(file);
#endif
RUN_TAIL(dump)
/* >8 mpi_dump_run */

/* ---------------------------------------------------------------- */

// /*#include argsused.h*/
// /*#include assused.h*/
DESTROY_HEAD(dump)
#if MPI
	if(s.runHere){ /*  No file to close if the device isn't active. */
		MPI_File_close(&(S->file));
	}
#else
  if (S->file) fclose(S->file); S->file=NULL;
#endif
DESTROY_TAIL(dump)

/*  ----------------------------------------------------------------// */

/* 8< mpi_dump_create */
CREATE_HEAD(dump)
	ACCEPTI(append,0,0,1);

#if MPI /*  Prepare for collective writing. */
	if (!accepts("file=",&(S->filename[0]),NULL,w)) return(0);
	
	/*  Communicator for active dump instances. */
	MPI_Comm comm_dump;
	
	/*  File */
	MPI_File file;
	int mode;
	int displacement=0;
	MPI_Offset filesize; /*  existing file size, for appending. */
	MPI_Aint prealloc_size;
	
	if(!deviceCommunicator(dev->s.runHere, &comm_dump)){
		ERROR("Could not create communicator.");
	}
		
	if(dev->s.runHere){ /*  Active devices only!	 */
		
		/*	Only define types on active instances,
		 *	as inactive instances will give invalid
		 *	subarray dimensions. */
		#include "dump_types.h"
		
		/*  Set access mode */
		if(S->append){
			mode = MPI_MODE_WRONLY | MPI_MODE_CREATE | MPI_MODE_APPEND;
		}else{
			mode = MPI_MODE_WRONLY | MPI_MODE_CREATE;
		}
		
		/*  Open file and set view */
		mpi_errno = MPI_File_open(comm_dump, S->filename, mode, MPI_INFO_NULL, &file);	
		
		/*  If requested by user, append to EOF */
		if(S->append){
			MPI_File_get_size(file, &filesize);
			displacement = (int)filesize;
		}
		
		mpi_errno = MPI_File_set_view(file, displacement, MPI_DOUBLE, filetype, "native", MPI_INFO_NULL);
		
		MPI_Type_extent(filetype,&prealloc_size);
		prealloc_size += displacement;
		MPI_File_preallocate(file,prealloc_size);
		
		/*  Assign the file to the parameter structure */
		S->file = file;
		S->sourceType = sourceType;
	}

#else /*  Sequential version */
	ACCEPTF(file,S->append?"ab":"wb",NULL);
#endif

CREATE_TAIL(dump,0)
/* >8 mpi_dump_create */
