/**
 *  Copyright (C) (2010-2012) Vadim Biktashev, Alexander Karpov, 
 *                            Irina Biktasheva, Ross McFarlane;
 *  Copyright (C) 2012 Sanjay Kharche
 *
 *  This file is part of Beatbox.
 *
 *   Beatbox is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Beatbox is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Beatbox.  If not, see <http://www.gnu.org/licenses/>.
 */


/* General pointwise ODE stepper */
/* The specific ODE rhs (e.g. cell model module) is attached as a separate function, or cell model module */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "system.h"
#include "beatbox.h"
#include "device.h"
#include "state.h"
#include "bikt.h"
#include "qpp.h"
#include "rhs.h"
#include "mpi_io_choice.h"

typedef rhs_str STR;

/****************/
RUN_HEAD(euler)
  DEVICE_CONST(real,ht)
  DEVICE_CONST(RhsProc *,f)
  DEVICE_CONST(Par,p)
  DEVICE_ARRAY(real,du)
  DEVICE_ARRAY(char,ode)
  DEVICE_CONST(Var,var)
  int x, y, z, v;
  real *u;

  for(x=s.x0;x<=s.x1;x++) {
    for(y=s.y0;y<=s.y1;y++) {
      for(z=s.z0;z<=s.z1;z++) {
	if(isTissue(x,y,z)){
	  u = (real *)(New + ind(x,y,z,s.v0));
	  if(!f(u,du,p,var,s.v1-s.v0+1)) {
	    URGENT_MESSAGE("error calculating %s at t=%ld point %d %d %d: u=",ode,t,x,y,z);
	    for(v=0;v<s.v1-s.v0+1;v++) URGENT_MESSAGE(" %lg",u[v]);
	    ERROR0("\n");
	    if(fflush_or_not==1)
	      fflush(stdout);
	  } /*  if !f... */
	  for(v=0;v<s.v1-s.v0+1;v++){
	    u[v] += ht*du[v]; // this is the actual calculation.
	    if (u[v]!=u[v]) {
	      URGENT_MESSAGE("NAN (not-a-number) detected at t=%ld x=%d y=%d z=%d v=%d\n",t,x,y,z,v);
	      return 0;
	    }
	  }
	} /*  if GEOMETRY_ON... */
      } /*  for z */
    } /*  for y */
  } /*  for x */
RUN_TAIL(euler)

/****************/
DESTROY_HEAD(euler)
  FREE(S->u);
  FREE(S->du);
  if (S->var.n) {
    FREE(S->var.src);
    FREE(S->var.dst);
  }
  FREE(S->p);
DESTROY_TAIL(euler)

/* Declare all available ODE's */
#define D(a) RhsProc a;
#include "rhslist.h"
#undef D
#define D(a) RhsCreate create_##a;
#include "rhslist.h"
#undef D

/****************************************/
CREATE_HEAD(euler) {

  int nv=dev->s.v1-dev->s.v0+1;
  int step, iv, ix, iy, iz, NV;

  ACCEPTR(ht,RNONE,0.,RNONE); /* ht=0 can make sense in some cases */
  ACCEPTS(ode,NULL);
  {
    char *pars;
    MALLOC(pars,(long)MAXSTRLEN);
    BEGINBLOCK("par=",pars);
    S->u=NULL;
    #define D(a)								\
    if (0==stricmp(S->ode,#a)) {					\
      NV=create_##a(&(S->p),&(S->var),pars,&(S->u),dev->s.v0);		\
      if (!NV) ERROR2("reading parameters for %s in \"%s\"",S->ode,pars); \
      if (NV!=nv) ERROR3("%d layers instead of %d for %s",nv,NV,S->ode); \
      S->f=a;								\
    } else
    #include "rhslist.h"
    #undef D
    ERROR1("unknown ODE %s",S->ode);
    ENDBLOCK;
    FREE(pars);
  }
  CALLOC(S->du,nv,sizeof(real));

  ACCEPTI(rest,0,0,INONE);
  if (S->rest) {
    if (S->u) MESSAGE("\n/* NOTICE: finding resting state while already defined by the rhs */");
    else CALLOC(S->u,nv,sizeof(real));
    if (S->var.n) MESSAGE("\n/* NOTICE: finding resting state while parameters are variable */");
    for(step=0;step<S->rest;step++) {
      if(!S->f(S->u,S->du,S->p,S->var,nv)) break;
      for(iv=0;iv<nv;iv++) (S->u)[iv]+=S->ht*S->du[iv];
    } /* for step */
  } /* if S->rest */

  if (S->u) {
    MESSAGE0("\n/* Resting state: "); for(iv=0;iv<nv;iv++) MESSAGE1("%lg ",(S->u)[iv]); MESSAGE0("*/");
    for (ix=dev->s.x0;ix<=dev->s.x1;ix++){
      for (iy=dev->s.y0;iy<=dev->s.y1;iy++){
	for (iz=dev->s.z0;iz<=dev->s.z1;iz++){
	  for (iv=dev->s.v0;iv<=dev->s.v1;iv++){
	    New[ind(ix,iy,iz,iv)]=(S->u)[iv-dev->s.v0];
	  } /* for iv */
	} /* for iz */
      } /* for iy */
    } /* for ix */
  } else {
    CALLOC(S->u,nv,sizeof(real));
  }

} CREATE_TAIL(euler,1)

