/**
 *  Copyright (C) (2010-2012) Vadim Biktashev, Alexander Karpov, 
 *                            Irina Biktasheva, Ross McFarlane;
 *  Copyright (C) 2012 Sanjay Kharche
 *
 *  This file is part of Beatbox.
 *
 *   Beatbox is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Beatbox is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Beatbox.  If not, see <http://www.gnu.org/licenses/>.
 */


/* 
 *	Carry out reduction operation over the space and assign result to a given k_variable.
 *	Available operations, selected in the user script are sum, product ('prod'), min and max.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>


#include "system.h"
#include "beatbox.h"
#include "error.h"
#include "device.h"
#include "state.h"
#include "qpp.h"

#define REDUCE(name) void name(real new_value, real *running)
typedef REDUCE(Reduce);

void reduce_max(real new_value, real *running){
  if(new_value > *running){
    *running = new_value;
  }
}

void reduce_min(real new_value, real *running){
  if(new_value < *running){
    *running = new_value;
  }
}

void reduce_sum(real new_value, real *running){
  *running += new_value;
}

void reduce_prod(real new_value, real *running){
  *running *= new_value;
}

typedef struct {
#if MPI
	MPI_Comm comm; 			// Communicator for reduction operation.
	int root; 				// Rank of device root in ALL_ACTIVE_PROCS
	MPI_Op global_reduce;	// Global reduction operation to use.
#endif
  char operation[10];		// Name of reduction operation to use.
  Reduce *local_reduce;	// Local reduction operation to use.
  double *result; 		// k_var to which result will be assigned.
  char *resultname;
  char debugname[MAXPATH];
  FILE *debug;
} STR;

/******************/
RUN_HEAD(reduce)
  double running = New[ind(s.x0,s.y0,s.z0,s.v0)];
  DEVICE_VAR(Reduce *, local_reduce)
  DEVICE_CONST(FILE *,debug)

#if MPI
  DEVICE_CONST(MPI_Comm, comm)
  DEVICE_CONST(int, root)
  DEVICE_CONST(MPI_Op, global_reduce)

  // Only active processes check their local max.
  if (s.runHere) {
#endif
    int x, y, z, v;
    int firstPass = 1;
    double this_value;
    for (x=s.x0;x<=s.x1;x++) {
      for (y=s.y0;y<=s.y1;y++) {
	for (z=s.z0;z<=s.z1;z++) {
	  for (v=s.v0;v<=s.v1;v++) {
	    this_value = New[ind(x,y,z,v)];
	    if (firstPass) {
	      running = this_value;
	      firstPass = 0;
	    } else {
	      (*local_reduce)(this_value, &running);
	    }
	  } /* for v */
	} /* for z */
      } /* for y */
    } /* for x */
#if MPI
    mpi_errno = MPI_Reduce(&running, S->result, 1, MPI_DOUBLE, global_reduce, 0, comm);
    CHECK_MPI_SUCCESS("Couldn't carry out reduction operation.")
    if (debug) fprintf(debug, "t=%ld : %s=%lf\n",t,S->resultname,*(S->result));
  } /* if(s.runHere) */

  /* ------------- DISTRIBUTE RESULT TO ALL PROCESSES ---------------*/
  MPI_Bcast(S->result, 1, MPI_DOUBLE, root, ALL_ACTIVE_PROCS);
#else
  *(S->result) = running;
#endif
RUN_TAIL(reduce)

/******************/
DESTROY_HEAD(reduce)
DESTROY_TAIL(reduce)

/******************/
CREATE_HEAD(reduce)
  DEVICE_ALWAYS_RUNS
  ACCEPTS(operation,"");
  if(strcmp(S->operation, "min")==0) {
    S->local_reduce = reduce_min;
#if MPI
    S->global_reduce = MPI_MIN;
#endif
  } else if(strcmp(S->operation, "max")==0) {
    S->local_reduce = reduce_max;
#if MPI
    S->global_reduce = MPI_MAX;
#endif
  } else if(strcmp(S->operation, "sum")==0) {
    S->local_reduce = reduce_sum;
#if MPI
    S->global_reduce = MPI_SUM;
#endif
  } else if(strcmp(S->operation, "prod")==0) {
    S->local_reduce = reduce_prod;
#if MPI
    S->global_reduce = MPI_PROD;
#endif
  } else {
    ERROR1("The given operation, \"%s\", was not recognised.\nAccepted values are \"min\", \"max\", \"sum\" and \"prod\".\n",
	   S->operation);
  }
  ACCEPTV(result);
  ACCEPTF(debug,"wt","");
#if MPI
  int device_rank, i_am_root, root, i;
  int rootFlags[num_active_procs];
  deviceCommunicator(dev->s.runHere, &(S->comm));
	
  /*	MPI_Reduce operates on the device communicator, defined where s.runHere is true.
   *	As the MPI_Broadcast that distributes the result must operate on ALL_ACTIVE_PROCS,
   *	it's necessary to identify the rank of the device root in ALL_ACTIVE_PROCS.
   *
   *	The MPI_Reduce root is 0 in the device communicator.
   *	The MPI_Broadcast root is computed here and stored in S->root.
   */

  i_am_root = 0;
  if(dev->s.runHere){
    MPI_Comm_rank(S->comm,&device_rank);
    if(device_rank == 0) i_am_root = 1;
  }
  mpi_errno = MPI_Allgather(&i_am_root, 1, MPI_INT, &rootFlags, 1, MPI_INT, ALL_ACTIVE_PROCS);
  CHECK_MPI_SUCCESS("Couldn't gather root flags.")
  for(i=0; i<num_active_procs; i++){
    if(rootFlags[i]) root = i;
  }
  S->root = root;
	
#endif
CREATE_TAIL(reduce,1)

